# Everything came from :
# http://docs.oracle.com/javase/specs/jvms/se7/html/jvms-4.html


import sys
import functools

import srddl.exceptions as se
import srddl.fields as sf
import srddl.helpers as sh
import srddl.models as sm

from srddl.core.offset import Size

# A feature to ask is to have a printing even in consol mode.
# An other feature request is to have an iterable for the sm.struct

# TODO List : 
# - Unified the ACC_ flag (there are different kind so..., at least do something
# clen)
# - handle the modified utf8 string in the Constant Pool
# - handle the long and double (the fact that they compt for double in the
# number of entry...)
# - made a more pretty thing about the printing of the float (for now it's
# considered as int...)
# - replacing all the get_size() function by a clean overriding of the ['size']
# - do something more clean for the structure (need more things is _srddl for 
# cleaning the structur with unknow size)
# - clean the bitfield
# - test everything (in particular the annotation)
# - replace somestuff by dictionnaries
# - clean the code...
# - do something more clean for the printer (srddl ?), with the linking with the
# constant table (for not having to look again at the beginning...)
# - seperate this shit in different file!
# - clean the problem of the unknow size for the buffer with something more
# clean than modifing it at run time
# - verifying the full compatibility SE7 and SE8, and possibly other stuff
# - integrate a desassembleur

Const_pool = None

class Constant_Pool(sm.Struct):
    cl_cp_tag = sf.IntField('Type entry', values=[
        # No value for 0
        sf.Value(1, 'CONSTANT_Utf8'),
        # No value for 2
        sf.Value(3, 'CONSTANT_Integer'),
        sf.Value(4, 'CONSTANT_Float'),
        sf.Value(5, 'CONSTANT_Long'),
        sf.Value(6, 'CONSTANT_Double'),
        sf.Value(7, 'CONSTANT_Class'),
        sf.Value(8, 'CONSTANT_String'),
        sf.Value(9, 'CONSTANT_FieldRef'),
        sf.Value(10, 'CONSTANT_MethodRef'),
        sf.Value(11, 'CONSTANT_InterfaceMethodRef'),
        sf.Value(12, 'CONSTANT_NameAndType'),
        # No value for 13, 14
        sf.Value(15, 'CONSTANT_MethodHandle'),
        sf.Value(16, 'CONSTANT_MethodType'),
        # No value for 17
        sf.Value(18, 'CONSTANT_InvokeDynamic')
    ])

    cp_type = None

    # TODO replace get_size by a clean overriding of the size
    # ask kushou about how to do it
    # TODO Do something more clean for the byte array field
    class CP_Utf8(sm.Struct):
        length = sf.IntField('Length', size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)

        # set a size of 1 and then modifiying in the _setup
        utf = sf.ByteArrayField(length, "UTF8-Modified")

        def get_size(self):
            return int(self.length['value']) + 2

        def pretty_printer(self, before='', end='\n'):
            print(before + str(self.length['value']) + " : " +
                    str(self.utf['value']), end=end)

    # TODO do something different for int and float with a correct pretty
    # printer for the float
    class CP_Float_Int(sm.Struct):
        value = sf.IntField('Value', size=sf.IntField.Size.INT32,
                endianess=sf.IntField.Endianess.BIG)

        def get_size(self):
            return self['size'].byte

        def pretty_printer(self, before='', end='\n'):
            print(before + str(self.value['value']), end=end)

    # TODO : test, this is wierd because the doc. say it take two entries, and
    # I'm not shure why they say that
    # TODO same as for CP_Float_Int but for the double instead of the float
    class CP_Double_Long(sm.Struct):
        low = sf.IntField('Low bytes', size=sf.IntField.Size.INT32,
                endianess=sf.IntField.Endianess.BIG)
        high = sf.IntField('High bytes', size=sf.IntField.Size.INT32,
                endianess=sf.IntField.Endianess.BIG)

        def get_size(self):
            return self['size'].byte

        def pretty_printer(self, before='', end='\n'):
            total = (self.high['value'] << 32) + self.low['value']
            print(before + str(total), end=end)

    class CP_Class(sm.Struct):
        name_index = sf.IntField('Name_Index', size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)

        def get_size(self):
            return self['size'].byte

        def pretty_printer(self, before='', end='\n'):
            print(before + str(self.name_index['value']), end=end)

    class CP_String(sm.Struct):
        string_index = sf.IntField('String_Index', size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)

        def get_size(self):
            return self['size'].byte

        def pretty_printer(self, before='', end='\n'):
            print(before + str(self.string_index['value']), end=end)

    class CP_Ref(sm.Struct):
        class_index = sf.IntField('Class Index', size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)
        name_type_index = sf.IntField('Name Type Index',
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)

        def get_size(self):
            return self['size'].byte

        def pretty_printer(self, before='', end='\n'):
            print(before + str(self.class_index['value']) + ", " +
                    str(self.name_type_index['value']), end=end)

    class CP_NameAndType(sm.Struct):
        name_index = sf.IntField('Name Index', size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)
        descriptor_index = sf.IntField('Descriptor Index',
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)

        def get_size(self):
            return self['size'].byte

        def pretty_printer(self, before='',end='\n'):
            print(before + str(self.name_index['value']) + ', ' +
                    str(self.descriptor_index['value']), end=end)

    class CP_MethodHandle(sm.Struct):
        reference_kind = sf.IntField('Kind', values =[
            sf.Value(1, 'REF_getField', 'getfield C.f:T'),
            sf.Value(2, 'REF_getStatic', 'getstatic C.f:T'),
            sf.Value(3, 'REF_putField', 'putfield C.f:T'),
            sf.Value(4, 'REF_putStatic', 'putstatic C.f:T'),
            sf.Value(5, 'REF_invokeVirtual', 'invokevirtual C.m:(A*)T'),
            sf.Value(6, 'REF_invokeStatic', 'invokestatic C.m:(A*)T'),
            sf.Value(7, 'REF_invokeSpecial', 'invokespecial C.m:(A*)T'),
            sf.Value(8, 'REF_newInvokeSpecial', 'new C; dup; invokespecial C.<init>:(A*)void'),
            sf.Value(9, 'REF_invokeInterface', 'invokeinterface C.m:(A*)T'),
        ])
        reference_index = sf.IntField('Reference Index',
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)

        def get_size(self):
            return self['size'].byte

        def pretty_printer(self, before='', end='\n'):
            print(before + str(self.reference_kind['name']) + "("
                    +str(self.reference_kind['value']) + "), " +
                    str(self.reference_index['value']), end=end)

    class CP_MethodType(sm.Struct):
        descriptor_index = sf.IntField('Descriptor Index',
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)

        def get_size(self):
            return self['size'].byte

        def pretty_printer(self, before='', end='\n'):
            print(before + str(self.descriptor_index['value']), end=end)

    class CP_InvokeDynamic(sm.Struct):
        # this must be a valid index into the bootstrap_methods array of the
        # bootstrap method table of this class file
        bootstrap_method_attr_index = sf.IntField('Bootstrap Method Attr Index',
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)
        name_and_type_index = sf.IntField('Name And Table Index',
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)

        def get_size(self):
            return self['size'].byte

        def pretty_printer(self, before='', end='\n'):
            print(before + str(self.bootstrap_method_attr_index['value']) + ', '
                    + str(self.name_and_type_index['value']))

    cp_type = None


    def _setup(self, data):
        if (self.cl_cp_tag['value'] == 1):
            self.cp_type = data.map(self._srddl._offset + 1, self.CP_Utf8)

        if (self.cl_cp_tag['value'] in (3, 4)):
            self.cp_type = data.map(self._srddl._offset + 1, self.CP_Float_Int)

        if (self.cl_cp_tag['value'] in (5, 6)):
            self.cp_type = data.map(self._srddl._offset + 1, self.CP_Double_Long)

        if (self.cl_cp_tag['value'] == 7):
            self.cp_type = data.map(self._srddl._offset + 1, self.CP_Class)

        if (self.cl_cp_tag['value'] == 8):
            self.cp_type = data.map(self._srddl._offset + 1, self.CP_String)

        if (self.cl_cp_tag['value'] in (9, 10, 11)):
            self.cp_type = data.map(self._srddl._offset + 1, self.CP_Ref)

        if (self.cl_cp_tag['value'] == 12):
            self.cp_type = data.map(self._srddl._offset + 1, self.CP_NameAndType)

        if (self.cl_cp_tag['value'] == 15):
            self.cp_type = data.map(self._srddl._offset + 1,
                    self.CP_MethodHandle)

        if (self.cl_cp_tag['value'] == 16):
            self.cp_type = data.map(self._srddl._offset + 1,
                    self.CP_MethodType)

        if (self.cl_cp_tag['value'] == 18):
            self.cp_type = data.map(self._srddl._offset + 1,
                    self.CP_InvokeDynamic)

    def pretty_printer(self, before='', end='\n'):
        print(before + str(self.cl_cp_tag['name']) + '(' +
                str(self.cl_cp_tag['value']) + ")", end='')
        if (self.cp_type != None):
            self.cp_type.pretty_printer(before=' ', end=end)

    # TODO override the real size method : _size
    def get_size(self):
        if self.cp_type != None:
            return 1 + self.cp_type.get_size()
        else:
            return 1


class class_java(sm.Struct):
    '''
    First Header :
        - magic
        - minor_version
        - major_version
        - constant_pool_count
    CONSTANT_POOL
    Next Header
    '''
    cl_magic = sf.ByteArrayField(4, 'Magic', valid=sh.equals(b'\xCA\xFE\xBA\xBE'))
    cl_minorv = sf.IntField('Minor Version', size=sf.IntField.Size.INT16,
            endianess=sf.IntField.Endianess.BIG)
    cl_majorv = sf.IntField('Major Version', size=sf.IntField.Size.INT16,
            endianess=sf.IntField.Endianess.BIG)
    # this major and minor version actually indicate which jdk are handling them
    # which is an interesting information but I don't know if I can put a
    # conditional things into the Value, it will be a good feature request, if
    # it's a number we match the number else if it's a function, this ont should
    # return a boolean, if True we match.
    cl_const_pool_count = sf.IntField('Size of the Constant Pool (+1)',
            size=sf.IntField.Size.INT16,
            endianess=sf.IntField.Endianess.BIG)

    def _setup(self, data):
        self.cl_cp = []
        done = 0
        for _ in range(int(self.cl_const_pool_count['value']) - 1):
            self.cl_cp.append(data.map(10 + done, Constant_Pool))
            done += self.cl_cp[-1].get_size()
        global Const_pool
        Const_pool = self.cl_cp
        self.next_header = data.map(10 + done, CL_NextHeader)

    def pretty_printer(self, before='', end='\n'):
        print(before + 'MAGIC : ' + str(self.cl_magic['value']), end=end)
        print(before + "MINORV : " + str(self.cl_minorv['value']), end=end)
        print(before + "MAJORV : " + str(self.cl_majorv['value']), end=end)
        print(before + "SIZE Constant Pool : "
                + str(int(self.cl_const_pool_count['value'])), end=end)

        if len(self.cl_cp) != 0:
            print(before + "CONSTANT POOL :",end=end)
        else:
            print(before + "CONSTANT POOL : (Empty)",end=end)

        for i in range(len(self.cl_cp)):
            self.cl_cp[i].pretty_printer(before=before + '\t #' + str(i + 1)
                    + ' ')

        self.next_header.pretty_printer()

class CL_NextHeader(sm.Struct):
    '''
    - access flag
    - this_class
    - super_class
    - interfaces_count
    INTERFACES
    '''
    cl_access_flag = sf.IntField('Access Flag', size=sf.IntField.Size.INT16,
            endianess=sf.IntField.Endianess.BIG, values=[]) # TODO something
    # more pretty than that there exist probably,  for now it will be handle in
    # the prettu_printer
    cl_this_class = sf.IntField('This Class', size=sf.IntField.Size.INT16,
            endianess=sf.IntField.Endianess.BIG)
    cl_super_class = sf.IntField('Super Class', size=sf.IntField.Size.INT16,
            endianess=sf.IntField.Endianess.BIG)
    cl_interface_count = sf.IntField('Interface Count', size=sf.IntField.Size.INT16,
            endianess=sf.IntField.Endianess.BIG)


    class CL_Interface(sm.Struct):
        """
        - just the interface number
        """
        cl_interface_number = sf.IntField('Interface Number',
                size=sf.IntField.Size.INT16, endianess=sf.IntField.Endianess.BIG)

    def _setup(self, data):
        self.cl_interfaces = []
        base_offset = self._srddl._offset + 8
        done = 0
        for _ in range(int(self.cl_interface_count['value'])):
            self.cl_interfaces.append(data.map(base_offset + done, self.CL_Interface))
            done += self.cl_interfaces[-1]['size']
        self.fields = data.map(base_offset + done, CL_Fields)


    def pretty_printer(self, before='', end='\n'):
        s = ""
        v = self.cl_access_flag['value']
        if v & 0x0001:
            s += ("" if s == "" else " | ") + "ACC_PUBLIC"
        if v & 0x0010:
            s += ("" if s == "" else " | ") + "ACC_FINAL"
        if v & 0x0020:
            s += ("" if s == "" else " | ") + "ACC_SUPER"
        if v & 0x0200:
            s += ("" if s == "" else " | ") + "ACC_INTERFACE"
        if v & 0x0400:
            s += ("" if s == "" else " | ") + "ACC_ABSTRACT"
        if v & 0x1000:
            s += ("" if s == "" else " | ") + "ACC_SYNTHETIC"
        if v & 0x2000:
            s += ("" if s == "" else " | ") + "ACC_ANNOTATION"
        if v & 0x4000:
            s += ("" if s == "" else " | ") + "ACC_ENUM"
        print(before + "ACCESS_FLAG : " + str(self.cl_access_flag['value'])
                + "(" + s + ")", end=end)
        print(before + "THIS_CLASS : " + str(self.cl_this_class['value']),
                end=end)
        print(before + "SUPER_CLASS : " + str(self.cl_super_class['value']),
                end=end)
        print(before + "INTERFACE_COUNT : "
                + str(self.cl_interface_count['value']), end=end)
        if len(self.cl_interfaces) == 0:
            print(before + "INTERFACE : (Empty)", end=end)
        else:
            print(before + "INTERFACE : ", end=end)
        for i in range(len(self.cl_interfaces)):
            print(before + "\t#" + str(i + 1) + " " +
                    str(self.cl_interfaces[i].cl_interface_number['value']),
                    end=end)
        self.fields.pretty_printer(before=before, end=end)

class CL_Fields(sm.Struct):
    """
    - fields count
    FIELDS
    """
    cl_field_count = sf.IntField('Field Count', size=sf.IntField.Size.INT16,
            endianess=sf.IntField.Endianess.BIG)

    class CL_Fields_Info(sm.Struct):
        access_flag = sf.IntField('Access Flag', size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)
        # TODO clean this with a real bitfield
        name_index = sf.IntField('Name Index', size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)
        descriptor_index = sf.IntField('Descriptor Index',
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)
        attributes_count = sf.IntField('Attributes Count',
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)

        def _setup(self, data):
            self.attributes = []
            base_offset = self._srddl._offset + 8
            done = 0
            for _ in range(int(self.attributes_count['value'])):
                self.attributes.append(data.map(base_offset + done, CL_Attribute))
                done += self.attributes[-1].get_size()

        def pretty_printer(self, before='', end='\n'):
            af = self.access_flag['value']
            s = ""
            if af & 0x0001:
                s += ("" if s == "" else " | ") + "ACC_PUBLIC"
            if af & 0x0002:
                s += ("" if s == "" else " | ") + "ACC_PRIVATE"
            if af & 0x0004:
                s += ("" if s == "" else " | ") + "ACC_PROTECTED"
            if af & 0x0008:
                s += ("" if s == "" else " | ") + "ACC_STATIC"
            if af & 0x0010:
                s += ("" if s == "" else " | ") + "ACC_FINAL"
            if af & 0x0040:
                s += ("" if s == "" else " | ") + "ACC_VOLATILE"
            if af & 0x0080:
                s += ("" if s == "" else " | ") + "ACC_TRANSIENT"
                # this is actually to look into for use-after-free
            if af & 0x1000:
                s += ("" if s == "" else " | ") + "ACC_SYNTHETIC"
            if af & 0x4000:
                s += ("" if s == "" else " | ") + "ACC_ENUM"

            print(before + "ACCESS_FLAG : " + str(af)
                    + " (" + s + ")", end=end)
            before += "\t"
            print(before + "NAME_INDEX : " + str(self.name_index['value']),
                    end=end)
            print(before + "DESCRIPTOR_INDEX : " +
                    str(self.descriptor_index['value']), end=end)
            print(before + "ATTRIBUTES_COUNT : " +
                    str(self.attributes_count['value']), end=end)

            if len(self.attributes) == 0:
                print(before + "ATTIBUTES : (Empty)", end=end)
            else:
                print(before + "ATTRIBUTES : ", end=end)
            for i in range(len(self.attributes)):
                self.attributes[i].pretty_printer(before=before + "\t#" + str(i) + " ", end=end)

        def get_size(self):
            n = 0
            for i in range(len(self.attributes)):
                n += self.attributes[i].get_size()
            return 8 + n


    def _setup(self, data):
        self.cl_fields = []
        base_offset = self._srddl._offset + 2
        done = 0
        for _ in range(int(self.cl_field_count['value'])):
            self.cl_fields.append(data.map(base_offset + done,
                self.CL_Fields_Info))
            done += self.cl_fields[-1].get_size()
        self.method = data.map(base_offset + done, CL_Method)


    def pretty_printer(self, before='', end='\n'):
        print(before + "FIELD_COUNT : " + str(self.cl_field_count['value']),
                end=end)
        if len(self.cl_fields) == 0:
            print(before + "FIELDS : (Empty)", end=end)
        else:
            print(before + "FIELDS : ", end=end)
        for i in range(len(self.cl_fields)):
            self.cl_fields[i].pretty_printer(before=before + "\t#" + str(i + 1)
            + " ", end=end)
        self.method.pretty_printer(before=before, end=end)

class CL_Method(sm.Struct):
    """
    - methods_count
    METHODS
    """
    cl_method_count = sf.IntField('Field Count', size=sf.IntField.Size.INT16,
            endianess=sf.IntField.Endianess.BIG)

    class CL_Method_Info(sm.Struct):
        """
            u2             access_flags;
            u2             name_index;
            u2             descriptor_index;
            u2             attributes_count;
            attribute_info attributes[attributes_count];
        """
        access_flags = sf.IntField('Access Flags', size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)
        name_index = sf.IntField('Name Index', size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)
        descriptor_index = sf.IntField('Descriptor Index',
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)
        attributes_count = sf.IntField('Attributes Count',
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)

        def _setup(self, data):
            self.attributes = []
            base_offset = self._srddl._offset + 8
            done = 0
            for _ in range(int(self.attributes_count['value'])):
                self.attributes.append(data.map(base_offset + done, CL_Attribute))
                done += self.attributes[-1].get_size()

        def pretty_printer(self, before='', end='\n'):
            af = self.access_flags['value']
            s = ""
            if af & 0x0001:
                s += ("" if s == "" else " | ") + "ACC_PUBLIC"
            if af & 0x0002:
                s += ("" if s == "" else " | ") + "ACC_PRIVATE"
            if af & 0x0004:
                s += ("" if s == "" else " | ") + "ACC_PROTECTED"
            if af & 0x0008:
                s += ("" if s == "" else " | ") + "ACC_STATIC"
            if af & 0x0010:
                s += ("" if s == "" else " | ") + "ACC_FINAL"
            if af & 0x0020:
                s += ("" if s == "" else " | ") + "ACC_SYNCHRONIZED"
            if af & 0x0040:
                s += ("" if s == "" else " | ") + "ACC_BRIDGE"
            if af & 0x0080:
                s += ("" if s == "" else " | ") + "ACC_VARARGS"
            if af & 0x0100:
                s += ("" if s == "" else " | ") + "ACC_NATIVE"
            if af & 0x0400:
                s += ("" if s == "" else " | ") + "ACC_ABSTRACT"
            if af & 0x0800:
                s += ("" if s == "" else " | ") + "ACC_STRICT"
            if af & 0x1000:
                s += ("" if s == "" else " | ") + "ACC_SYNTHETIC"

            print(before + "ACCESS_FLAGS : " + str(af)
                    + " (" + s + ")", end=end)
            before += "\t"
            print(before + "NAME_INDEX : " + str(self.name_index['value']),
                    end=end)
            print(before + "DESCRIPTOR_INDEX : " +
                    str(self.descriptor_index['value']), end=end)
            print(before + "ATTRIBUTES_COUNT : " +
                    str(self.attributes_count['value']), end=end)

            if len(self.attributes) == 0:
                print(before + "ATTIBUTES : (Empty)", end=end)
            else:
                print(before + "ATTRIBUTES : ", end=end)
            for i in range(len(self.attributes)):
                self.attributes[i].pretty_printer(before=before + "\t#" + str(i) + " ", end=end)

        def get_size(self):
            n = 0
            for i in range(len(self.attributes)):
                n += self.attributes[i].get_size()
            return 8 + n

    def _setup(self, data):
        self.cl_methods = []
        base_offset = self._srddl._offset + 2
        done = 0
        for _ in range(int(self.cl_method_count['value'])):
            self.cl_methods.append(data.map(base_offset + done,
                self.CL_Method_Info))
            done += self.cl_methods[-1].get_size()
        self.cl_attributes = data.map(base_offset + done, CL_Class_Attributes)

    def pretty_printer(self, before='', end='\n'):
        print(before + "METHOD_COUNT : " + str(self.cl_method_count['value']),
                end=end)
        if len(self.cl_methods) == 0:
            print(before + "METHODS : (Empty)", end=end)
        else:
            print(before + "METHODS : ", end=end)
        for i in range(len(self.cl_methods)):
            self.cl_methods[i].pretty_printer(before=before + "\t#" + str(i + 1)
            + " ", end=end)
        self.cl_attributes.pretty_printer(before=before, end=end)

class CL_Class_Attributes(sm.Struct):
    """
        u2             attributes_count;
        attribute_info attributes[attributes_count];
    """
    cl_attributes_count = sf.IntField('Attributes Count',
            size=sf.IntField.Size.INT16,
            endianess=sf.IntField.Endianess.BIG)

    def _setup(self, data):
        self.cl_attributes = []
        base_offset = self._srddl._offset + 2
        done = 0
        for _ in range(int(self.cl_attributes_count['value'])):
            self.cl_attributes.append(data.map(base_offset + done, CL_Attribute))
            done += self.cl_attributes[-1].get_size()

    def pretty_printer(self, before='', end='\n'):
        print(before + "ATTRIBUTES_COUNT : " +
                str(self.cl_attributes_count['value']), end=end)
        for i in range(len(self.cl_attributes)):
            self.cl_attributes[i].pretty_printer(before=before + "\t#"
                    + str(i + 1) + " ", end=end)

class CL_Exception(sm.Struct):
    """
       u2 start_pc;
       u2 end_pc;
       u2 handler_pc;
       u2 catch_type;
    """
    start_pc = sf.IntField('Start PC',
            size=sf.IntField.Size.INT16,
            endianess=sf.IntField.Endianess.BIG)
    end_pc = sf.IntField('End PC',
            size=sf.IntField.Size.INT16,
            endianess=sf.IntField.Endianess.BIG)
    handler_pc = sf.IntField('Handler PC',
            size=sf.IntField.Size.INT16,
            endianess=sf.IntField.Endianess.BIG)
    catch_type = sf.IntField('Catch Type',
            size=sf.IntField.Size.INT16,
            endianess=sf.IntField.Endianess.BIG)

    def pretty_printer(self, before='', end='\n'):
        print(before + "START_PC : " + str(self.start_pc['value']), end=end)
        before += '\t'
        print(before + "END_PC : " + str(self.end_pc['value']), end=end)
        print(before + "HANDLER_PC : " + str(self.handler_pc['value']), end=end)
        print(before + "CATCH_TYPE : " + str(self.catch_type['value']), end=end)

    def get_size(self):
        return 4

# Needed for the StackFrameTable In the CL_Attribute
class VerificationTypeInfo_Internal(sm.Struct):
    """
        union verification_type_info {
            Top_variable_info;
            Integer_variable_info;
            Float_variable_info;
            Long_variable_info;
            Double_variable_info;
            Null_variable_info;
            UninitializedThis_variable_info;
            Object_variable_info;
            Uninitialized_variable_info;
        }
    """

    tag = sf.IntField("Tag")

    class ObjectVariable_Internal(sm.Struct):
        cpool_index = sf.IntField("Cpool Index",
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)

        def get_size(self):
            return 2

        def pretty_printer(self, before='', end='\n'):
            print(before + "OBJ_VAR_INFO_CPOOL_INDEX : "
                    + str(self.cpool_index['value']), end=end)

    class UninitializedVariable_Internal(sm.Struct):
        offset = sf.IntField("Offset",
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)

        def get_size(self):
            return 2

        def pretty_printer(self, before='', end='\n'):
            print(before + "UNINITIALIZED_VAR_INFO_OFFSET : "
                    + str(self.offset['value']), end=end)

    def _setup(self, data):
        if self.tag['value'] == 7:
            self.complement = data.map(self._srddl._offset + 1,
                    self.ObjectVariable_Internal)
        elif self.tag['value'] == 8:
            self.complement = data.map(self._srddl._offset + 1,
                    self.UninitializedVariable_Internal)
        else:
            self.complement = None


    def get_size(self):
        if self.tag['value'] in (7, 8):
            return 3
        else:
            return 1

    def pretty_printer(self, before='', end='\n'):
        d = {
            0: "ITEM_Top",
            1: "ITEM_Integer",
            2: "ITEM_Float",
            3: "ITEM_Double",
            4: "ITEM_Long",
            5: "ITEM_Null",
            6: "ITEM_UninitializedThis",
            7: "ITEM_Object",
            8: "ITEM_Uninitialized"
        }

        print(before + "VERIFICATION_TYPE_INFO TAG : "
                + str(self.tag['value']) + " ("
                + d[int(self.tag['value'])] + ")", end=end)
        if self.tag['value'] in (7, 8):
            self.complement.pretty_printer(before=before + '\t',
                    end=end)

class CL_Attribute(sm.Struct): # attribute_info struct
    """
    - attribute_name_index
    - attribute_length
    - info[attribute_length]
    The info can contain a lot of things including the code
    """ # TODO parse this and put more detailed information about what is in
    #every type of Attributes
    cl_attribute_name_index = sf.IntField('Attribute Name Index',
            size=sf.IntField.Size.INT16,
            endianess=sf.IntField.Endianess.BIG)
    cl_attribute_length = sf.IntField('Attribute Length',
            size=sf.IntField.Size.INT32,
            endianess=sf.IntField.Endianess.BIG)
    f = lambda s: s.cl_attribute_length['value'] if (
            Const_pool[s.cl_attribute_name_index['value']
                - 1].cp_type.utf['value'] == b"SourceDebugExtension") else 0
    cl_info = sf.ByteArrayField(f, "Info of the Attribute")

    def _setup(self, data):
        #self.cl_info._size = Size() + 0

        base_offset = self._srddl._offset + 6
        ani = self.cl_attribute_name_index['value'] - 1
        self.cl_type = None
        # TODO do this with a dictionnary it would have be really more clean 
        if (Const_pool[ani].cp_type.__class__ == Constant_Pool.CP_Utf8):
            if Const_pool[ani].cp_type.utf['value'] == b'ConstantValue':
                self.cl_type = data.map(base_offset,
                        self.CL_Attribute_ConstantValue)
            elif Const_pool[ani].cp_type.utf['value'] == b'Code':
                self.cl_type = data.map(base_offset,
                        self.CL_Attribute_Code)
            elif Const_pool[ani].cp_type.utf['value'] == b'Exceptions':
                self.cl_type = data.map(base_offset,
                        self.CL_Attribute_Exception)
            elif Const_pool[ani].cp_type.utf['value'] == b'InnerClasses':
                self.cl_type = data.map(base_offset,
                        self.CL_Attribute_InnerClasses)
            elif Const_pool[ani].cp_type.utf['value'] == b'EnclosingMethod':
                self.cl_type = data.map(base_offset,
                        self.CL_Attribute_EnclosingMethod)
            elif Const_pool[ani].cp_type.utf['value'] == b'Synthetic':
                self.cl_type = data.map(base_offset,
                        self.CL_Attribute_Synthetic)
            elif Const_pool[ani].cp_type.utf['value'] == b'Signature':
                self.cl_type = data.map(base_offset,
                        self.CL_Attribute_Signature)
            elif Const_pool[ani].cp_type.utf['value'] == b'SourceFile':
                self.cl_type = data.map(base_offset,
                        self.CL_Attribute_SourceFile)
            #elif Const_pool[ani].cp_type.utf['value'] == b'SourceDebugExtension':
            #    self.cl_info._size = (Size() # TODO test this!
            #        + int(self.cl_attribute_length['value']))
            elif Const_pool[ani].cp_type.utf['value'] == b'LineNumberTable':
                self.cl_type = data.map(base_offset,
                        self.CL_Attribute_LineNumberTable)
            elif Const_pool[ani].cp_type.utf['value'] == b'LocalVariableTable':
                self.cl_type = data.map(base_offset,
                        self.CL_Attribute_LocalVariableTable)
            elif (Const_pool[ani].cp_type.utf['value'] ==
                    b'LocalVariableTypeTable'):
                self.cl_type = data.map(base_offset,
                        self.CL_Attribute_LocalVariableTypeTable)
            elif Const_pool[ani].cp_type.utf['value'] == b'Deprecated':
                self.cl_type = data.map(base_offset,
                        self.CL_Attribute_Deprecated)
            elif Const_pool[ani].cp_type.utf['value'] == b'BootstrapMethods':
                self.cl_type = data.map(base_offset,
                        self.CL_Attribute_BootstrapMethods)
            elif Const_pool[ani].cp_type.utf['value'] == b'MethodParameters':
                self.cl_type = data.map(base_offset,
                        self.CL_Attribute_MethodParameters)
            elif Const_pool[ani].cp_type.utf['value'] == b'StackMapTable':
                self.cl_type = data.map(base_offset,
                        self.CL_Attribute_StackMapTable)
            elif Const_pool[ani].cp_type.utf['value'] == b'RuntimeVisibleAnnotations':
                self.cl_type = data.map(base_offset,
                        self.CL_Attribute_RuntimeVisibleAnnotations)
            elif Const_pool[ani].cp_type.utf['value'] == b'RuntimeInvisibleAnnotations':
                self.cl_type = data.map(base_offset,
                        self.CL_Attribute_RuntimeInvisibleAnnotations)
            elif Const_pool[ani].cp_type.utf['value'] == b'RuntimeVisibleParameterAnnotations':
                self.cl_type = data.map(base_offset,
                        self.CL_Attribute_RuntimeVisibleParameterAnnotations)
            elif Const_pool[ani].cp_type.utf['value'] == b'RuntimeInvisibleParameterAnnotations':
                self.cl_type = data.map(base_offset,
                        self.CL_Attribute_RuntimeInvisibleParameterAnnotations)
            elif Const_pool[ani].cp_type.utf['value'] == b'RuntimeInvisibleTypeAnnotations':
                self.cl_type = data.map(base_offset,
                        self.CL_Attribute_RuntimeInvisibleTypeAnnotations)
            elif Const_pool[ani].cp_type.utf['value'] == b'RuntimeVisibleTypeAnnotations':
                self.cl_type = data.map(base_offset,
                        self.CL_Attribute_RuntimeVisibleTypeAnnotations)
            elif Const_pool[ani].cp_type.utf['value'] == b'AnnotationDefault':
                self.cl_type = data.map(base_offset,
                        self.CL_Attribute_AnnotationDefault)
            else:
                print(str(Const_pool[ani].cp_type.utf['value']))
            #print(Const_pool[ani].cp_type.utf['value'])


    def pretty_printer(self, before='', end='\n'):
        print(before + "NAME_INDEX : " +
                str(self.cl_attribute_name_index['value']), end=end)
        before += "\t"
        print(before + "ATTRIBUTE_LENGTH : " +
                str(self.cl_attribute_length['value']), end=end)
        #print(before + "INFO : " + str(self.cl_info['value']), end=end)
        if self.cl_type != None:
            self.cl_type.pretty_printer(before=before, end=end)
        else:
            if self.cl_info._size.byte != 0:
                print(before + "SourceDebugExtension : "
                        + str(self.cl_info['value']), end=end)
            else:
                print(before + "ATTRIBUTE NOT DEFINED", end=end)

    def get_size(self):
        return 6 + int(self.cl_attribute_length['value'])

    class CL_Attribute_ConstantValue(sm.Struct):
        constant_value_index = sf.IntField('Constant Value Index',
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)

        def pretty_printer(self, before='', end='\n'):
            print(before + "CONST VALUE : " +
                    str(self.constant_value_index['value']), end=end)

        def get_size(self):
            return 2

    # TODO desassemble the code
    class CL_Attribute_Code(sm.Struct):
        max_stack = sf.IntField('Max Stack', size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)
        max_locals = sf.IntField('Max Locals', size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)
        code_length = sf.IntField('Code Length', size=sf.IntField.Size.INT32,
                endianess=sf.IntField.Endianess.BIG)
        code = sf.ByteArrayField(code_length, "Info of the Attribute")

        def _setup(self, data):
            off = self._srddl._offset + 8 + int(self.code_length['value'])
            self.exception = data.map(off, self.CL_Excep)

        def get_size(self):
            return (8 + int(self.code_length['value'])
                    + self.exception.get_size())

        def pretty_printer(self, before='', end='\n'):
            print(before + "MAX_STACK : " + str(self.max_stack['value']),
                    end=end)
            print(before + "MAX_LOCALS : " + str(self.max_locals['value']),
                    end=end)
            print(before + "CODE_LENGTH : " + str(self.code_length['value']),
                    end=end)
            print(before + "CODE : " + str(self.code['value']), end=end)
            self.exception.pretty_printer(before=before, end=end)

        class CL_Excep(sm.Struct):
            exception_table_length = sf.IntField('Size table',
                    size=sf.IntField.Size.INT16,
                    endianess=sf.IntField.Endianess.BIG)

            def _setup(self, data):
                self.exception_table = []
                base_offset = self._srddl._offset + 2
                done = 0
                for _ in range(int(self.exception_table_length['value'])):
                    self.exception_table.append(data.map(base_offset + done,
                        CL_Exception))
                    done += 8
                self.attributes = data.map(base_offset + done,
                        CL_Class_Attributes)

            def pretty_printer(self, before='', end='\n'):
                print(before + "EXCEPTION_TABLE_LENGTH : " +
                        str(self.exception_table_length['value']), end=end)
                if len(self.exception_table) == 0:
                    print(before + "EXCEPTION TABLE : (Empty)", end=end)
                else:
                    print(before + "EXCEPTION TABLE : ", end=end)
                for i in range(len(self.exception_table)):
                    self.exception_table[i].pretty_printer(before=before + "\t#"
                            + str(i) + " ", end=end)
                self.attributes.pretty_printer(before=before, end=end)

            def get_size(self):
                return 2 + 8 * int(self.exception_table_length['value'])


    class CL_Attribute_Exception(sm.Struct):
        """
            u2 number_of_exceptions;
            u2 exception_index_table[number_of_exceptions];
        """

        number_of_exceptions = sf.IntField("Number Of Exceptions",
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)

        class Exception_Internal(sm.Struct):
            exception_index = sf.IntField("Exception Index",
                    size=sf.IntField.Size.INT16,
                    endianess=sf.IntField.Endianess.BIG)

            def get_size(self):
                return 2
            def pretty_printer(self, before='', end='\n'):
                print(before + str(self.exception_index['value']), end=end)

        def _setup(self, data):
            self.exception_index_table = []
            base_offset = self._srddl._offset + 2
            done = 0
            for _ in range(int(self.number_of_exceptions['value'])):
                self.exception_index_table.append(data.map(base_offset + done,
                    self.Exception_Internal))
                done += 2

        def get_size(self):
            return 2 + len(self.exception_index_table) * 2

        def pretty_printer(self, before='', end='\n'):
            print(before + "NUMBER EXCEPTIONS : " +
                    str(self.number_of_exceptions['value']), end=end)
            if len(self.exception_index_table) == 0:
                print(before + "EXCEPTIONS INDEX TABLE : (Empty)", end=end)
            else:
                print(before + "EXCEPTIONS INDEX TABLE :", end=end)
            for i in range(len(self.exception_index_table)):
                self.exception_index_table[i].pretty_printer(before=before +
                        "\t#" + str(i) + " ", end=end)

    class CL_Attribute_InnerClasses(sm.Struct):
        """
            u2 number_of_classes;
            {
                u2 inner_class_info_index;
                u2 outer_class_info_index;
                u2 inner_name_index;
                u2 inner_class_access_flags;
            } classes[number_of_classes];
        """
        number_of_classes = sf.IntField("Number of InnerClasses",
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)

        class InnerClasses_Internal(sm.Struct):
            inner_class_info_index = sf.IntField("InnerClass Info Index",
                    size = sf.IntField.Size.INT16,
                    endianess = sf.IntField.Endianess.BIG)
            outer_class_info_index = sf.IntField("OuterClass Info Index",
                    size = sf.IntField.Size.INT16,
                    endianess = sf.IntField.Endianess.BIG)
            inner_name_index = sf.IntField("InnerClass Name Index",
                    size = sf.IntField.Size.INT16,
                    endianess = sf.IntField.Endianess.BIG)
            inner_class_access_flags = sf.IntField("InnerClass Access Flags",
                    size = sf.IntField.Size.INT16,
                    endianess = sf.IntField.Endianess.BIG)

            def get_size(self):
                return 8

            def pretty_printer(self, before='', end='\n'):
                print(before + "INNERCLASS Index : " +
                        str(self.inner_class_info_index['value']), end=end)
                before += "\t"
                print(before + "OUTERCLASS Index : " +
                        str(self.inner_class_info_index['value']), end=end)
                if self.inner_name_index['value'] == 0:
                    print(before + "INNERCLASS Name Index : 0 (Anonymous)",
                            end=end)
                else:
                    print(before + "INNERCLASS Name Index : " + 
                            str(self.inner_name_index['value']), end=end)

                f = self.inner_class_access_flags['value']
                s = ""
                if f & 0x0001:
                    s += ("" if s == "" else " | ") + "ACC_PUBLIC"
                if f & 0x0002:
                    s += ("" if s == "" else " | ") + "ACC_PRIVATE"
                if f & 0x0004:
                    s += ("" if s == "" else " | ") + "ACC_PROTECTED"
                if f & 0x0008:
                    s += ("" if s == "" else " | ") + "ACC_STATIC"
                if f & 0x0010:
                    s += ("" if s == "" else " | ") + "ACC_FINAL"
                if f & 0x0200:
                    s += ("" if s == "" else " | ") + "ACC_INTERFACE"
                if f & 0x0400:
                    s += ("" if s == "" else " | ") + "ACC_ABSTRACT"
                if f & 0x1000:
                    s += ("" if s == "" else " | ") + "ACC_SYNTHETIC"
                if f & 0x2000:
                    s += ("" if s == "" else " | ") + "ACC_ANNOTATION"
                if f & 0x4000:
                    s += ("" if s == "" else " | ") + "ACC_ENUM"
                if s != "":
                    s = " (" + s + ")"
                print(before + "INNERCLASS Access Flags : " +
                        str(f) + s , end=end)

        def _setup(self, data):
            self.classes = []
            done = 0
            base_offset = self._srddl._offset + 2
            for _ in range(int(self.number_of_classes['value'])):
                self.classes.append(data.map(base_offset + done,
                    self.InnerClasses_Internal))
                done += 8

        def get_size(self):
            return 2 + 8 * int(self.number_of_classes['value'])

        def pretty_printer(self, before='', end='\n'):
            print(before + "NUMBER_INNERCLASSES : " +
                    str(self.number_of_classes['value']), end=end)
            if self.classes == []:
                print(before + "INNERCLASSES : (Empty)", end=end)
            else:
                print(before + "INNERCLASSES :", end=end)
            for i in range(len(self.classes)):
                self.classes[i].pretty_printer(before=before + "\t#" + str(i) +
                        " ", end=end)

    class CL_Attribute_EnclosingMethod(sm.Struct):
        """
            u2 class_index;
            u2 method_index;
        """
        class_index = sf.IntField("Class Index", size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)
        method_index = sf.IntField("Method Index", size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)

        def get_size(self):
            return 4

        def pretty_printer(self, before="", end="\n"):
            print(before + "ENCLOSINGMETHOD CLASS : "
                    + str(self.class_index['value']), end=end)
            print(before + "EMCLOSINGMETHOD METHOD : "
                    + str(self.method_index['value']), end=end)

    class CL_Attribute_Synthetic(sm.Struct):
        """
            This class is actually to be empty
        """
        # TODO test!

        def get_size(self):
            return 0

        def pretty_printer(self, before="", end="\n"):
            print(before + "ATTRIBUTE_SYNTHETIC", end=end)

    class CL_Attribute_Signature(sm.Struct):
        """
            u2 signature_index;
        """
        signature_index = sf.IntField("Signature Index",
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)

        def get_size(self):
            return 2

        def pretty_printer(self, before="", end="\n"):
            print(before + "SIGNATURE INDEX : " +
                    str(self.signature_index['value']), end=end)

    class CL_Attribute_SourceFile(sm.Struct):
        """
            u2 sourcefile_index;
        """
        sourcefile_index = sf.IntField("Source File Index",
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)

        def get_size(self):
            return 2

        def pretty_printer(self, before="", end="\n"):
            print(before + "SOURCE FILE INDEX : " +
                    str(self.sourcefile_index['value']), end=end)

    class CL_Attribute_LineNumberTable(sm.Struct):
        """
            u2 line_number_table_length;
            {
                u2 start_pc;
                u2 line_number; 
            } line_number_table[line_number_table_length];
        """
        line_number_table_length = sf.IntField("Line Number Table Length",
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)

        class LineNumber_Internal(sm.Struct):
            start_pc = sf.IntField("Start PC",
                    size=sf.IntField.Size.INT16,
                    endianess=sf.IntField.Endianess.BIG)
            line_number = sf.IntField("Line Number",
                    size=sf.IntField.Size.INT16,
                    endianess=sf.IntField.Endianess.BIG)

            def get_size(self):
                return 4

            def pretty_printer(self, before='', end="\n"):
                print(before + "START PC : " + str(self.start_pc['value']),
                        end=end)
                print(before + "\tLINE NUMBER : "
                        + str(self.line_number['value']), end=end)



        def _setup(self, data):
            self.line_number_table = []
            base_offset = self._srddl._offset
            done = 0
            for _ in range(int(self.line_number_table_length['value'])):
                self.line_number_table.append(data.map(base_offset + done,
                    self.LineNumber_Internal))
                done += 2

        def get_size(self):
            return 2 + self.line_number_table_length['value'] * 4
        
        def pretty_printer(self, before="", end="\n"):
            print(before + "LINE_NUMBER TABLE LENGTH : " +
                    str(self.line_number_table_length['value']), end=end)
            if len(self.line_number_table) == 0:
                print(before + "LINE_NUMBER TABLE : (Empty)", end=end)
            else:
                print(before + "LINE_NUMBER TABLE :", end=end)
            for i in range(len(self.line_number_table)):
                self.line_number_table[i].pretty_printer(before=before + "\t#" +
                        str(i) + " ", end=end)


    # TODO test
    class CL_Attribute_LocalVariableTable(sm.Struct):
        """
            u2 local_variable_table_length;
            {
                u2 start_pc;
                u2 length;
                u2 name_index;
                u2 descriptor_index;
                u2 index;
            } local_variable_table[local_variable_table_length];
        """
        local_variable_table_length = sf.IntField("Local Variable Table Length",
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)

        class LocalVariable_Internal(sm.Struct):
            """
                u2 start_pc;
                u2 length;
                u2 name_index;
                u2 descriptor_index;
                u2 index;
            """
            start_pc = sf.IntField("Start PC", size=sf.IntField.Size.INT16,
                    endianess=sf.IntField.Endianess.BIG)
            length = sf.IntField("Length", size=sf.IntField.Size.INT16,
                    endianess=sf.IntField.Endianess.BIG)
            name_index = sf.IntField("Name Index", size=sf.IntField.Size.INT16,
                    endianess=sf.IntField.Endianess.BIG)
            descriptor_index = sf.IntField("Descriptor Index",
                    size=sf.IntField.Size.INT16,
                    endianess=sf.IntField.Endianess.BIG)
            index = sf.IntField("Index", size=sf.IntField.Size.INT16,
                    endianess=sf.IntField.Endianess.BIG)

            def get_size(self):
                return 10

            def pretty_printer(self, before='', end='\n'):
                print(before + "START_PC : " + str(self.start_pc['value']),
                        end=end)
                before += "\t"
                print(before + "LENGTH : " + str(self.length['value']), end=end)
                print(before + "NAME_INDEX : " + str(self.name_index['value']),
                        end=end)
                print(before + "DESCRIPTOR_INDEX : " +
                        str(self.descriptor_index['value']), end=end)
                print(before + "INDEX : " + str(self.index['value']), end=end)

        def _setup(self, data):
            self.local_variable_table = []
            offset = self._srddl._offset + 2
            for _ in range(int(self.local_variable_table_length['value'])):
                self.local_variable_table.append(data.map(offset,
                    self.LocalVariable_Internal))
                offset += 10

        def get_size(self):
            return 2 + 10 * int(self.local_variable_table_length['value'])

        def pretty_printer(self, before='', end='\n'):
            print(before + "LOCAL_VARIABLE_TABLE_LENGTH : "
                    + str(self.local_variable_table_length['value']), end=end)
            if self.local_variable_table == []:
                print(before + "LOCAL_VARIABLE_TABLE : (Empty)", end=end)
            else:
                print(before + "LOCAL_VARIABLE_TABLE :", end=end)

            for i in len(self.local_variable_table):
                self.local_variable_table[i].pretty_printer(before=before
                        + "\t#" + str(i) + " ", end=end)

    # TODO test
    class CL_Attribute_LocalVariableTypeTable(sm.Struct):
        """
            u2 local_variable_type_table_length;
            {
                u2 start_pc;
                u2 length;
                u2 name_index;
                u2 signature_index;
                u2 index;
            } local_variable_type_table[local_variable_type_table_length];
        """
        local_variable_type_table_length = sf.IntField(
                "LOCAL_VARIABLE_TYPE_TABLE",
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)

        class LocalVariableTypeTable_Internal(sm.Struct):
            """
                u2 start_pc;
                u2 length;
                u2 name_index;
                u2 signature_index;
                u2 index;
            """
            start_pc = sf.IntField("Start PC", size=sf.IntField.Size.INT16,
                    endianess=sf.IntField.Endianess.BIG)
            length = sf.IntField("Length", size=sf.IntField.Size.INT16,
                    endianess=sf.IntField.Endianess.BIG)
            name_index = sf.IntField("Name Index", size=sf.IntField.Size.INT16,
                    endianess=sf.IntField.Endianess.BIG)
            signature_index = sf.IntField("Signature Index",
                    size=sf.IntField.Size.INT16,
                    endianess=sf.IntField.Endianess.BIG)
            index = sf.IntField("Index", size=sf.IntField.Size.INT16,
                    endianess=sf.IntField.Endianess.BIG)

            def get_size(self):
                return 10

            def pretty_printer(self, before='', end='\n'):
                print(before + "START_PC : " + str(self.start_pc['value']),
                        end=end)
                before += "\t"
                print(before + "LENGTH : " + str(self.length['value']),
                        end=end)
                print(before + "NAME_INDEX : " + str(self.name_index['value']),
                        end=end)
                print(before + "SIGNATURE_INDEX : "
                        + str(self.signature_index['value']), end=end)
                print(before + "INDEX : " + str(self.index['value']),
                        end=end)

        def _setup(self, data):
            off = 2 + self._srddl._offset
            self.local_var_type_table = []
            for _ in range(int(self.local_variable_type_table_length['value'])):
                self.local_var_type_table.append(data.map(off,
                    self.LocalVariableTypeTable_Internal))
                off += 10
        
        def get_size(self):
            return 2 + 10 * int(self.local_variable_type_table_length['value'])

        def pretty_printer(self, before='', end='\n'):
            print(before + "LOCAL_VAR_TYPE_TABLE_LEN : "
                    + str(self.local_variable_type_table_length['value']),
                    end=end)
            for i in range(len(self.local_var_type_table)):
                self.local_var_type_table[i].pretty_printer(before=before
                        + "\t#" + str(i) + " ", end=end)

    # TODO test
    class CL_Attribute_Deprecated(sm.Struct):
        """
            This class is actually to be empty
        """

        def get_size(self):
            return 0

        def pretty_printer(self, before="", end="\n"):
            print(before + "ATTRIBUTE_DEPRECATED", end=end)

    # TODO Annotations

    # TODO test, really not sure if this work...
    class CL_Attribute_BootstrapMethods(sm.Struct):
        """
            u2 num_bootstrap_methods;
            {   u2 bootstrap_method_ref;
                u2 num_bootstrap_arguments;
                u2 bootstrap_arguments[num_bootstrap_arguments];
            } bootstrap_methods[num_bootstrap_methods];
        """
        num_bootstrap_methods = sf.IntField("Number of Bootstrap Methods",
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)

        class BootstrapMethod_Internal(sm.Struct):
            bootstrap_method_ref = sf.IntField("Bootstrap Method Ref",
                    size=sf.IntField.Size.INT16,
                    endianess=sf.IntField.Endianess.BIG)
            num_bootstrap_arg = sf.IntField("Number of Bootstrap Argument",
                    size=sf.IntField.Size.INT16,
                    endianess=sf.IntField.Endianess.BIG)

            class ArgBootstrap_Internal(sm.Struct):
                bootstrap_arg = sf.IntField("Bootstrap Argument",
                        size=sf.IntField.Size.INT16,
                        endianess=sf.IntField.Endianess.BIG)

                def get_size(self):
                    return 2
                def pretty_printer(self, before='', end='\n'):
                    print(before + self.bootstrap_arg['value'], end=end)

            def _setup(self, data):
                self.bootstrap_arguments = []
                off = self._srddl._offset + 4
                for _ in range(int(self.num_bootstrap_arg['value'])):
                    self.bootstrap_arguments.append(data.map(off,
                        self.ArgBootstrap_Internal))
                    off += 2

            def get_size(self):
                return 4 + 2 * int(self.num_bootstrap_arg['value'])

            def pretty_printer(self, before='', end='\n'):
                print(before + "BOOTSTRAP_METHOD_REF : " +
                        str(self.bootstrap_method_ref['value']), end=end)
                before += "\t"
                print(before + "NUMBER_BOOTSTRAP_ARGS : " +
                        str(self.num_bootstrap_arg['value']), end=end)
                if len(self.bootstrap_arguments) == 0:
                    print(before + "BOOTSTRAP_ARGS : (Empty)", end=end)
                else:
                    print(before + "BOOTSTRAP_ARGS :", end=end)
                for i in self.bootstrap_arguments:
                    self.bootstrap_arguments[i].pretty_printer(before=before
                            + "\t#" + str(i) + " ", end=end)

        def _setup(self, data):
            self.bootstrap_methods = []
            off = self._srddl._offset + 2
            for _ in range(int(self.num_bootstrap_methods['value'])):
                self.bootstrap_methods.append(data.map(off,
                    self.BootstrapMethod_Internal))
                off += self.bootstrap_methods[i].get_size()

        def get_size(self):
            s = 2
            for i in len(self.bootstrap_methods):
                s += self.bootstrap_methods[i].get_size()
            return s

        def pretty_printer(self, before='', end='\n'):
            print(before + "NUMBER_BOOTSTRAP_METHODS : " +
                    str(self.num_bootstrap_methods['value']), end=end)
            if self.bootstrap_methods == []:
                print(before + "BOOTSTRAP_METHODS : (Empty)", end=end)
            else:
                print(before + "BOOTSTRAP_METHODS :", end=end)
            for i in range(len(self.bootstrap_methods)):
                self.bootstrap_methods.pretty_printer(before=before + "\t#" +
                        str(i) + " ", end=end)

    class CL_Attribute_MethodParameters(sm.Struct):
        """
            u1 parameters_count;
            {
                u2 name_index;
                u2 access_flags;
            } parameters[parameters_count];
        """
        parameters_count = sf.IntField("Parameters Count",
                endianess=sf.IntField.Endianess.BIG)

        class Parameters_Internal(sm.Struct):
            name_index = sf.IntField("Name Index",
                    size=sf.IntField.Size.INT16,
                    endianess=sf.IntField.Endianess.BIG)
            access_flags = sf.IntField("Access Flags",
                    size=sf.IntField.Size.INT16,
                    endianess=sf.IntField.Endianess.BIG)

            def get_size(self):
                return 4

            def pretty_printer(self, before='', end='\n'):
                print(before + "NAME INDEX : " + str(self.name_index['value']),
                        end=end)
                before += "\t"
                s = ""
                f = self.access_flags['value']
                if f & 0x10:
                    s += ("" if s == "" else " | ") + "ACC_FINAL"
                if f & 0x1000:
                    s += ("" if s == "" else " | ") + "ACC_SYNTHETIC"
                if f & 0x8000:
                    s += ("" if s == "" else " | ") + "ACC_MANDATED"
                if s != "":
                    s = " (" + s + ")"
                print(before + "ACCESS_FLAGS : "
                        + str(self.access_flags['value']) + s, end=end)

        def _setup(self, data):
            off = self._srddl._offset + 1
            self.parameters = []
            for _ in range(len(int(self.parameters_count['value']))):
                self.parameters.append(data.map(off, self.Parameters_Internal))
                off += 4

        def get_size(self):
            return 1 + 4 * int(self.parameters_count['value'])

        def pretty_printer(self, before='', end='\n'):
            print(before + "PARAMETERS_COUNT : " +
                    str(self.parameters_count['value']), end=end)
            if len(self.parameters) == 0:
                print(before + "PARAMETERS : (Empty)", end=end)
            else:
                print(before + "PARAMETERS :", end=end)
            for i in range(len(self.parameters)):
                self.parameters[i].pretty_printer(before=before + "\t#" + str(i)
                        + " ", end=end)



    class CL_Attribute_StackMapTable(sm.Struct):
        """
            u2              number_of_entries;
            stack_map_frame entries[number_of_entries];
        """
        number_of_entries = sf.IntField("Number of entries",
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)

        class StackMapFrame_Internal(sm.Struct):
            """
                union stack_map_frame {
                    same_frame;
                    same_locals_1_stack_item_frame;
                    same_locals_1_stack_item_frame_extended;
                    chop_frame;
                    same_frame_extended;
                    append_frame;
                    full_frame;
                }
            """
            frame_type = sf.IntField("Frame Type") # on 1 byte only

            class Local_one_stack_item_ext_Internal(sm.Struct):
                """
                    u2 offset_delta;
                    verification_type_info stack[1];
                """
                offset_delta = sf.IntField("Offset Delta",
                        size=sf.IntField.Size.INT16,
                        endianess=sf.IntField.Endianess.BIG)

                def _setup(self, data):
                    # I could use a "Super struct" but I don't know how to
                    # use them TODO
                    self.stack = data.map(self._srddl._offset + 2,
                            VerificationTypeInfo_Internal)

                def get_size(self):
                    return 2 + self.stack.get_size()

                def pretty_printer(self, before='', end='\n'):
                    print(before + "OFFSET_DELTA : " +
                            str(self.offset_delta['value']), end=end)
                    self.stack.pretty_printer(before=before, end=end)

            class Chopframe_Internal(sm.Struct):
                """
                    u2 offset_delta;
                """
                offset_delta = sf.IntField("Offset Delta",
                        size=sf.IntField.Size.INT16,
                        endianess=sf.IntField.Endianess.BIG)

                def get_size(self):
                    return 2

                def pretty_printer(self, before='', end='\n'):
                    print(before + "OFFSET_DELTA : " +
                            str(self.offset_delta['value']), end=end)

            class FullFrame_Internal(sm.Struct):
                """
                    u2 offset_delta;
                    u2 number_of_locals;
                    verification_type_info locals[number_of_locals];
                    u2 number_of_stack_items;
                    verification_type_info stack[number_of_stack_items];
                """

                offset_delta = sf.IntField("Offset Delta",
                        size=sf.IntField.Size.INT16,
                        endianess=sf.IntField.Endianess.BIG)
                number_of_locals = sf.IntField("Number locals",
                        size=sf.IntField.Size.INT16,
                        endianess=sf.IntField.Endianess.BIG)

                class FullFrame_Stack_Internal(sm.Struct):

                    number_of_stack_item = sf.IntField("Number Stack Item",
                        size=sf.IntField.Size.INT16,
                        endianess=sf.IntField.Endianess.BIG)

                    def _setup(self, data):
                        off = self._srddl._offset + 2
                        self.stack = []
                        for _ in range(int(self.number_of_stack_item['value'])):
                            self.stack.append(data.map(off,
                                VerificationTypeInfo_Internal))
                            off += self.stack[-1].get_size()

                    def get_size(self):
                        i = 2
                        for e in self.stack:
                            i += e.get_size()
                        return i

                    def pretty_printer(self, before='', end='\n'):
                        print(before + "NUMBER_STACK_ITEM : " +
                                str(self.number_of_stack_item['value']),
                                end=end)
                        if len(self.stack) == 0:
                            print(before + "STACK_ITEMS : (Empty)", end=end)
                        else:
                            print(before + "STACK_ITEMS :", end=end)
                        for i in range(len(self.stack)):
                            self.stack[i].pretty_printer(before=before + "\t#" +
                                    str(i) + " ", end=end)

                def _setup(self, data):
                    off = self._srddl._offset + 4
                    self.local = []
                    for _ in range(int(self.number_of_locals['value'])):
                        self.local.append(data.map(off,
                            VerificationTypeInfo_Internal))
                        off += self.local[-1].get_size()
                    self.stack = data.map(off, self.FullFrame_Stack_Internal)

                def get_size(self):
                    i = 4
                    for e in self.local:
                        i += e.get_size()
                    return i + self.stack.get_size()

                def pretty_printer(self, before='', end='\n'):
                    print(before + "OFFSET_DELTA : " +
                            str(self.offset_delta['value']), end=end)
                    print(before + "NUMBER_OF_LOCALS : " +
                            str(self.number_of_locals['value']), end=end)
                    if self.local == []:
                        print(before + "LOCALS : (Empty)", end=end)
                    else:
                        print(before + "LOCALS :", end=end)
                    for i in range(len(self.local)):
                        self.local[i].pretty_printer(before=before + "\t#" +
                                str(i) + " ", end=end)
                    self.stack.pretty_printer(before=before, end=end)

            class Append_Internal(sm.Struct):
                frame_type = sf.IntField("Frame Type") # on 1 byte only
                offset_delta = sf.IntField("Offset Delta",
                        size=sf.IntField.Size.INT16,
                        endianess=sf.IntField.Endianess.BIG)

                def _setup(self, data):
                    self.local = []
                    off = self._srddl._offset + 3
                    for i in range(int(self.frame_type['value']) - 251):
                        self.local.append(data.map(off,
                            VerificationTypeInfo_Internal))
                        off += self.local[-1].get_size()

                def get_size(self):
                    i = 2
                    for e in self.local:
                        i += e.get_size()
                    return i

                def pretty_printer(self, before='', end='\n'):
                    print(before + "OFFSET_DELTA : " +
                            str(self.offset_delta['value']), end=end)
                    if self.local == []:
                        print(before + "LOCALS : (Empty)", end=end)
                    else:
                        print(before + "LOCALS :", end=end)
                    for i in range(len(self.local)):
                        self.local[i].pretty_printer(before=before + "\t#" +
                                str(i) + " ", end=end)

            # TODO pu the VerificationTypeInfo_Internal class in a more suited
            # place.
            def _setup(self, data):
                base_offset = self._srddl._offset + 1
                if self.frame_type['value'] in range(64): # 0-63
                    self.typ = 0
                elif self.frame_type['value'] in range(64, 128): # 64-127
                    self.typ = 1
                    self.stack = data.map(base_offset,
                            VerificationTypeInfo_Internal)
                elif self.frame_type['value'] == 247:
                    self.typ = 2
                    self.struct = data.map(base_offset,
                            self.Local_one_stack_item_ext_Internal)
                elif self.frame_type['value'] in range(248, 251): # 248 - 250
                    self.typ = 3
                    self.struct = data.map(base_offset, self.Chopframe_Internal)
                elif self.frame_type['value'] == 251:
                    self.typ = 4
                    # I use the same struct as the one before, but there are not
                    # the same in reality (even if they have the same content)
                    self.struct = data.map(base_offset, self.Chopframe_Internal)
                elif self.frame_type['value'] in range(252, 255): # 252 - 254
                    self.typ = 5
                    # Overlapping works! but that's not really clean...
                    # TODO do something clean
                    self.struct = data.map(base_offset - 1,
                            self.Append_Internal)
                elif self.frame_type['value'] == 255:
                    self.typ = 6
                    self.struct = data.map(base_offset, self.FullFrame_Internal)

            def get_size(self):
                if self.typ == 0:
                    return 1
                elif self.typ == 1:
                    return 1 + self.stack.get_size()
                else:
                    return 1 + self.struct.get_size()

            def pretty_printer(self, before='', end='\n'):
                d = {
                    0:"SAME",
                    1:"SAME_LOCALS_1_STACK_ITEM",
                    2:"SAME_LOCALS_1_STACK_ITEM_EXTENDED",
                    3:"CHOP",
                    4:"SAME_FRAME_EXTENDED",
                    5:"APPEND",
                    6:"FULL_FRAME",
                }
                if self.typ == 1:
                    self.stack.pretty_printer(before=before, end=end)
                elif self.typ not in (0, 1, 7):
                    self.struct.pretty_printer(before=before, end=end)


    #class StackMapFrame_Internal(sm.Struct):
    #number_of_entries = sf.IntField("Number of entries",
        def _setup(self, data):
            self.entries = []
            off = self._srddl._offset + 2
            for _ in range(int(self.number_of_entries['value'])):
                self.entries.append(data.map(off, self.StackMapFrame_Internal))
                off += self.entries[-1].get_size()

        def get_size(self):
            i = 2
            for e in self.entries:
                i += e.get_size()
            return i

        def pretty_printer(self, before='', end='\n'):
            print(before + "STACK_MAP_TABLES_SIZE_TABLE : " +
                    str(self.number_of_entries['value']), end=end)
            if self.entries == []:
                print(before + "ENTRIES : (Empty)", end=end)
            else:
                print(before + "ENTRIES :", end=end)
            for i in range(len(self.entries)):
                self.entries[i].pretty_printer(before=before + "\t#" + str(i)
                        + " ", end=end)

    class CL_Attribute_RuntimeVisibleAnnotations(sm.Struct):
        """
            u2         attribute_name_index;
            u4         attribute_length;
            u2         num_annotations;
            annotation annotations[num_annotations];
        """
        attribute_name_index = sf.IntField("Attribute Name Index",
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)
        attribute_length = sf.IntField("Attribute Length",
                size=sf.IntField.Size.INT32,
                endianess=sf.IntField.Endianess.BIG)
        num_annotations = sf.IntField("Num Annotations",
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)

        def _setup(self, data):
            self.annotations = []
            off = self._srddl._offset + 8
            for _ in range(int(self.num_annotations['value'])):
                self.annotations.append(data.map(off, CL_Annotation))
                off += self.annotations[-1].get_size()

        def get_size(self):
            i = 8
            for e in self.annotations:
                i += e.get_size()
            return i

        def pretty_printer(self, before='', end='\n'):
            print(before + "ATTRIBUTE_NAME_INDEX : " +
                    str(self.attribute_name_index['value']), end=end)
            before += "\t"
            print(before + "ATTRIBUTE_LENGTH  : " +
                    str(self.attribute_length['value']), end=end)
            print(before + "NUM_ANNOTATIONS : " +
                    str(self.num_annotations['value']), end=end)
            if len(self.annotations) == 0:
                print(before + "ANNOTATIONS : (Empty)", end=end)
            else:
                print(before + "ANNOTATIONS :", end=end)
            for i in range(len(self.annotations)):
                self.annotations[i].pretty_printer(before=before + "\t#" +
                        str(i) + " ", end=end)

    class CL_Attribute_RuntimeInvisibleAnnotations(sm.Struct):
        """
            u2         attribute_name_index;
            u4         attribute_length;
            u2         num_annotations;
            annotation annotations[num_annotations];
        """
        attribute_name_index = sf.IntField("Attribute Name Index",
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)
        attribute_length = sf.IntField("Attribute Length",
                size=sf.IntField.Size.INT32,
                endianess=sf.IntField.Endianess.BIG)
        num_annotations = sf.IntField("Num Annotations",
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)

        def _setup(self, data):
            self.annotations = []
            off = self._srddl._offset + 8
            for _ in range(int(self.num_annotations['value'])):
                self.annotations.append(data.map(off, CL_Annotation))
                off += self.annotations[-1].get_size()

        def get_size(self):
            i = 8
            for e in self.annotations:
                i += e.get_size()
            return i

        def pretty_printer(self, before='', end='\n'):
            print(before + "ATTRIBUTE_NAME_INDEX : " +
                    str(self.attribute_name_index['value']), end=end)
            before += "\t"
            print(before + "ATTRIBUTE_LENGTH  : " +
                    str(self.attribute_length['value']), end=end)
            print(before + "NUM_ANNOTATIONS : " +
                    str(self.num_annotations['value']), end=end)
            if len(self.annotations) == 0:
                print(before + "ANNOTATIONS : (Empty)", end=end)
            else:
                print(before + "ANNOTATIONS :", end=end)
            for i in range(len(self.annotations)):
                self.annotations[i].pretty_printer(before=before + "\t#" +
                        str(i) + " ", end=end)

    class CL_Attribute_RuntimeVisibleParameterAnnotations(sm.Struct):
        """
            u2 attribute_name_index;
            u4 attribute_length;
            u1 num_parameters;
            {   u2         num_annotations;
                annotation annotations[num_annotations];
            } parameter_annotations[num_parameters];
        """
        attribute_name_index = sf.IntField("Attribute Name Index",
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)
        attribute_length = sf.IntField("Attribute Length",
                size=sf.IntField.Size.INT32,
                endianess=sf.IntField.Endianess.BIG)
        num_parameters = sf.IntField("Num Parameters")

        class Param_Anno_Internal(sm.Struct):
            num_annotations = sf.IntField("Num Annotations",
                    size=sf.IntField.Size.INT16,
                    endianess=sf.IntField.Endianess.BIG)

            def _setup(self, data):
                self.annotations = []
                off = self._srddl._offset + 2
                for _ in range(int(self.num_annotations['value'])):
                    self.annotations.append(data.map(off, CL_Annotation))
                    off += self.annotations[-1].get_size()

            def get_size(self):
                i = 2
                for _ in range(len(self.annotations)):
                    i += self.annotations[i].get_size()
                return i

            def pretty__printer(self, before='', end='\n'):
                print(before + "NUM_ANNOTATIONS : " +
                        str(self.num_annotations['value']), end=end)
                if len(self.annotations) == 0:
                    print(before + "ANNOTATIONS : (Empty)", end=end)
                else:
                    print(before + "ANNOTATIONS :", end=end)
                for i in range(len(self.annotations)):
                    self.annotations[i].pretty_printer(before=before + "\t#" +
                            str(i) + " ", end=end)

        def _setup(self, data):
            self.parameter_annotations = []
            off = self._srddl._offset + 7
            for _ in range(int(self.num_parameters['value'])):
                self.parameter_annotations.append(data.map(off,
                    self.Param_Anno_Internal))
                off += self.parameter_annotations[-1].get_size()

        def get_size(self):
            i = 7
            for e in self.parameter_annotations:
                i += e.get_size()
            return i

        def pretty_printer(self, before='', end='\n'):
            print(before + "ATTRIBUTE_NAME_INDEX : " +
                    str(self.attribute_name_index['value']), end=end)
            before += "\t"
            print(before + "ATTRIBUTE_LENGTH  : " +
                    str(self.attribute_length['value']), end=end)
            print(before + "NUM_PARAMETERS : " +
                    str(self.num_parameters['value']), end=end)
            if len(self.parameter_annotations) == 0:
                print(before + "PARAMETERS : (Empty)", end=end)
            else:
                print(before + "PARAMETERS :", end=end)
            for i in range(len(self.parameter_annotations)):
                self.parameter_annotations[i].pretty_printer(before=before
                        + "\t#" + str(i) + " ", end=end)

    class CL_Attribute_RuntimeInvisibleParameterAnnotations(sm.Struct):
        """
            u2 attribute_name_index;
            u4 attribute_length;
            u1 num_parameters;
            {   u2         num_annotations;
                annotation annotations[num_annotations];
            } parameter_annotations[num_parameters];
        """
        attribute_name_index = sf.IntField("Attribute Name Index",
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)
        attribute_length = sf.IntField("Attribute Length",
                size=sf.IntField.Size.INT32,
                endianess=sf.IntField.Endianess.BIG)
        num_parameters = sf.IntField("Num Parameters")

        class Param_Anno_Internal(sm.Struct):
            num_annotations = sf.IntField("Num Annotations",
                    size=sf.IntField.Size.INT16,
                    endianess=sf.IntField.Endianess.BIG)

            def _setup(self, data):
                self.annotations = []
                off = self._srddl._offset + 2
                for _ in range(int(self.num_annotations['value'])):
                    self.annotations.append(data.map(off, CL_Annotation))
                    off += self.annotations[-1].get_size()

            def get_size(self):
                i = 2
                for _ in range(len(self.annotations)):
                    i += self.annotations[i].get_size()
                return i

            def pretty__printer(self, before='', end='\n'):
                print(before + "NUM_ANNOTATIONS : " +
                        str(self.num_annotations['value']), end=end)
                if len(self.annotations) == 0:
                    print(before + "ANNOTATIONS : (Empty)", end=end)
                else:
                    print(before + "ANNOTATIONS :", end=end)
                for i in range(len(self.annotations)):
                    self.annotations[i].pretty_printer(before=before + "\t#" +
                            str(i) + " ", end=end)

        def _setup(self, data):
            self.parameter_annotations = []
            off = self._srddl._offset + 7
            for _ in range(int(self.num_parameters['value'])):
                self.parameter_annotations.append(data.map(off,
                    self.Param_Anno_Internal))
                off += self.parameter_annotations[-1].get_size()

        def get_size(self):
            i = 7
            for e in self.parameter_annotations:
                i += e.get_size()
            return i

        def pretty_printer(self, before='', end='\n'):
            print(before + "ATTRIBUTE_NAME_INDEX : " +
                    str(self.attribute_name_index['value']), end=end)
            before += "\t"
            print(before + "ATTRIBUTE_LENGTH  : " +
                    str(self.attribute_length['value']), end=end)
            print(before + "NUM_PARAMETERS : " +
                    str(self.num_parameters['value']), end=end)
            if len(self.parameter_annotations) == 0:
                print(before + "PARAMETERS : (Empty)", end=end)
            else:
                print(before + "PARAMETERS :", end=end)
            for i in range(len(self.parameter_annotations)):
                self.parameter_annotations[i].pretty_printer(before=before
                        + "\t#" + str(i) + " ", end=end)

    class CL_Attribute_RuntimeVisibleTypeAnnotations(sm.Struct):
        """
            u2              attribute_name_index;
            u4              attribute_length;
            u2              num_annotations;
            type_annotation annotations[num_annotations];
        """
        attribute_name_index = sf.IntField("Attribute Name Index",
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)
        attribute_length = sf.IntField("Attribute Length",
                size=sf.IntField.Size.INT32,
                endianess=sf.IntField.Endianess.BIG)
        num_annotations = sf.IntField("Number Annotations",
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)

        def _setup(self, data):
            off = self._srddl._offset + 8
            self.annotations = []
            for _ in range(int(self.num_annotations['value'])):
                self.annotations.append(data.map(off, Type_annotation))

        def get_size(self):
            i = 8
            for e in self.annotations:
                i += e.get_size()
            return i

        def pretty_printer(self, before='', end='\n'):
            print(before + "ATTRIBUTE_NAME_INDEX : " +
                    str(self.attribute_name_index['value']), end=end)
            before += "\t"
            print(before + "ATTRIBUTE_LENGTH  : " +
                    str(self.attribute_length['value']), end=end)
            print(before + "NUM_ANNOTATIONS : " +
                    str(self.num_annotations['value']), end=end)
            if len(self.annotations) == 0:
                print(before + "ANNOTATIONS_TYPE : (Empty)", end=end)
            else:
                print(before + "ANNOTATIONS_TYPE :", end=end)
            for i in range(len(self.parameter_annotations)):
                self.annotations[i].pretty_printer(before=before
                        + "\t#" + str(i) + " ", end=end)

    class CL_Attribute_RuntimeInvisibleTypeAnnotations(sm.Struct):
        """
            u2              attribute_name_index;
            u4              attribute_length;
            u2              num_annotations;
            type_annotation annotations[num_annotations];
        """
        attribute_name_index = sf.IntField("Attribute Name Index",
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)
        attribute_length = sf.IntField("Attribute Length",
                size=sf.IntField.Size.INT32,
                endianess=sf.IntField.Endianess.BIG)
        num_annotations = sf.IntField("Number Annotations",
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)

        def _setup(self, data):
            off = self._srddl._offset + 8
            self.annotations = []
            for _ in range(int(self.num_annotations['value'])):
                self.annotations.append(data.map(off, Type_annotation))

        def get_size(self):
            i = 8
            for e in self.annotations:
                i += e.get_size()
            return i

        def pretty_printer(self, before='', end='\n'):
            print(before + "ATTRIBUTE_NAME_INDEX : " +
                    str(self.attribute_name_index['value']), end=end)
            before += "\t"
            print(before + "ATTRIBUTE_LENGTH  : " +
                    str(self.attribute_length['value']), end=end)
            print(before + "NUM_ANNOTATIONS : " +
                    str(self.num_annotations['value']), end=end)
            if len(self.annotations) == 0:
                print(before + "ANNOTATIONS_TYPE : (Empty)", end=end)
            else:
                print(before + "ANNOTATIONS_TYPE :", end=end)
            for i in range(len(self.parameter_annotations)):
                self.annotations[i].pretty_printer(before=before
                        + "\t#" + str(i) + " ", end=end)

    class CL_Attribute_AnnotationDefault(sm.Struct):
        """
            u2            attribute_name_index;
            u4            attribute_length;
            element_value default_value;
        """
        attribute_name_index = sf.IntField("Attribute Name Index",
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)
        attribute_length = sf.IntField("Attribute Length",
                size=sf.IntField.Size.INT32,
                endianess=sf.IntField.Endianess.BIG)

        def _setup(self, data):
            self.default_value = data.map(self._srddl._offset + 6,
                    Element_value)
        def get_size(self):
            return 6 + self.default_value.get_size()
        def pretty_printer(self, before='', end='\n'):
            print(before + "ATTRIBUTE_NAME_INDEX : " +
                    str(self.attribute_name_index['value']), end=end)
            before += "\t"
            print(before + "ATTRIBUTE_LENGTH  : " +
                    str(self.attribute_length['value']), end=end)
            print(before + "DEFAULT_VALUE :", end=end)
            self.default_value.pretty_printer(before=before + "\t", end=end)


# TODO test
class Type_annotation(sm.Struct):
    target_type = sf.IntField("Target Type") # on one byte

    class Type_Parameter_Target(sm.Struct):
        type_parameter_index = sf.IntField("Type Parameter Index")
        def get_size(self):
            return 1
        def pretty_printer(self, before='', end='\n'):
            print(before + "TYPE_PARAMETER_INDEX : "
                    + str(self.type_parameter_index['value']), end=end)

    class SuperType_Target(sm.Struct):
        supertype_index = sf.IntField("SuperType Index",
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)
        def get_size(self):
            return 2
        def pretty_printer(self, before='', end='\n'):
            print(before + "SUPERTYPE_INDEX : " + 
                    str(self.supertype_index['value']), end=end)

    class TypeParameter_BoundTarget(sm.Struct):
        type_parameter_index = sf.IntField("Type Parameter Index")
        bound_index = sf.IntField("Bount Index")
        def get_size(self):
            return 2
        def pretty_printer(self, before='', end='\n'):
            print(before + "TYPE_PARAMETER_INDEX : " +
                    str(self.type_parameter_index['value']), end=end)
            print(before + "\tBOUND_INDEX : " + str(self.bound_index['value']),
                    end=end)

    class Formal_Parameter_Type(sm.Struct):
        formal_parameter_index = sf.IntField("Formal Parameter Index")
        def get_size(self):
            return 1
        def pretty_printer(self, before='', end='\n'):
            print(before + "FORMAL_PARAMETER_INDEX : " +
                    str(self.formal_paramter_index['value']), end=end)

    class Throws_Target(sm.Struct):
        throws_type_index = sf.IntField("Throw Type Index",
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)
        def get_size(self):
            return 2
        def pretty_printer(self, before='', end='\n'):
            print(before + "THROWS_TYPE_INDEX : "
                    + str(self.throws_type_index['value']), end=end)

    class Local_Var_Target(sm.Struct):
        table_length = sf.IntField("Table Length",
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)
        class Table_Internal(sm.Struct):
            start_pc = sf.IntField("Start Pc",
                    size=sf.IntField.Size.INT16,
                    endianess=sf.IntField.Endianess.BIG)
            length = sf.IntField("Length",
                    size=sf.IntField.Size.INT16,
                    endianess=sf.IntField.Endianess.BIG)
            index = sf.IntField("Index",
                    size=sf.IntField.Size.INT16,
                    endianess=sf.IntField.Endianess.BIG)
            def get_size(self):
                return 6
            def pretty_printer(self, before='', end='\n'):
                print(before + "START_PC : " + str(self.start_pc['value']),
                        end=end)
                before += '\t'
                print(before + "LENGTH : " + str(self.length['value']), end=end)
                print(before + "INDEX : " + str(self.index['value']), end=end)
        def _setup(self, data):
            off = self._srddl._offset + 2
            self.table = []
            for _ in range(int(self.table_length['value'])):
                self.table.append(data.map(off, self.Table_Internal))
                off += self.table[-1].get_size()
        def get_size(self):
            return 2 + self.table_length['value'] * 6
        def pretty_printer(self, before='', end='\n'):
            print(before + "TABLE_LENGTH : " + str(self.table_length['value']),
                    end=end)
            if self.table == []:
                print(before + "TABLE : (Empty)", end=end)
            else:
                print(before + "TABLE :", end=end)
            for i in range(len(self.table)):
                self.table[i].pretty_printer(before=before + "\t#" + str(i)
                + " ", end=end)

    class Catch_Target(sm.Struct):
        exception_table_index = sf.IntField("Exception Table Index",
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)
        def get_size(self):
            return 2
        def pretty_printer(self, before='', end='\n'):
            print(before + "EXCEPTION_TABLE_INDEX : " +
                    str(self.exception_table_index['value']), end=end)

    class Offset_Target(sm.Struct):
        offset = sf.IntField("Offset",
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)
        def get_size(self):
            return 2
        def pretty_printer(self, before='', end='\n'):
            print(before + "OFFSET : " + str(self.offset['value']), end=end)

    class Type_Argument_Target(sm.Struct):
        offset = sf.IntField("Offset",
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)
        type_arg_index = sf.IntField("Type Argument Index")
        def get_size(self):
            return 3
        def pretty_printer(self, before='', end='\n'):
            print(before + "OFFSET : " + str(self.offset['value']), end=end)
            print(before + "\tTYPE_ARG_INDEX : "
                    + str(self.type_arg_index['value']), end=end)

    class Type_Path(sm.Struct):
        path_length = sf.IntField("Path Length")
        class Path_Internal(sm.Struct):
            type_path_kind = sf.IntField("Type Path Kind")
            type_argument_index = sf.IntField("Type Argument Index")
            def get_size(self):
                return 2
            def pretty_printer(self, before='', end='\n'):
                print(before + "TYPE_PATH_KIND : " +
                        str(self.type_path_kind['value']), end=end)
                print(before + "TYPE_ARGYMENT_INDEX : " +
                        str(self.type_argument_index['value']), end=end)
        def _setup(self, data):
            self.path = []
            off = self._srddl._offset + 1
            for _ in range(int(self.path_length['value'])):
                self.path.append(data.map(off, self.Path_Internal))
                off += 2
        def get_size(self):
            return 1 + 2 * self.path_length['value']
        def pretty_printer(self, before='', end='\n'):
            print(before + "PATH_LENGTH : " + str(self.path_length['value']),
                    end=end)
            if self.path == []:
                print(before + "PATH : (Empty)", end=end)
            else:
                print(before + "PATH :", end=end)
            for i in range(len(self.path)):
                self.path[i].pretty_printer(before=before + "\t#" + str(i)
                        + " ", end=end)

    class Type_Annot_End_Internal(sm.Struct):
        type_index = sf.IntField("Type Index",
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)
        num_elt_value_pairs = sf.IntField("Num Elt Value Pairs", 
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)
        class Elt_Value_Pairs(sm.Struct):
            element_name_index = sf.IntField("Elt Name Index",
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)
            def _setup(self, data):
                self.value = data.map(self._srddl._offset + 2, Element_value)
            def get_size(self):
                return 2 + self.value.get_size()
            def pretty_printer(self, before='', end='\n'):
                print(before + "ELEMENT_NAME_INDEX : " +
                        str(self.element_name_index['value']), end=end)
                self.value.pretty_printer(before=before+"\t", end=end)
        def _setup(self, data):
            self.elt_value_pairs = []
            off = self._srddl._offset + 4
            for _ in range(int(self.num_elt_value_pairs)):
                self.elt_value_pairs.append(data.map(off, self.Elt_Value_Pairs))
                off += self.elt_value_pairs[-1].get_size()
        def get_size(self):
            i = 4
            for e in self.elt_value_pairs:
                i += e.get_size()
            return i
        def pretty_printer(self, before='', end='\n'):
            print(before + "TYPE_INDEX : " + str(self.type_index['value']),
                    end=end)
            print(before + "NUM_ELT_VALUE_PAIRS : " +
                    str(self.num_elt_value_pairs['value']), end=end)
            if self.elt_value_pairs == []:
                print(before + "ELT_VALUE_PAIRS : (Empty)", end=end)
            else:
                print(before + "ELT_VALUE_PAIRS :", end=end)
            for i in range(len(self.elt_value_pairs)):
                self.elt_value_pairs[i].pretty_printer(before=before + "\t#" +
                        str(i) + " ", end=end)

    def _setup(self, data):
        self.target_info = None
        off = self._srddl._offset + 1
        if self.target_type['value'] in (0x00, 0x01):
            self.target_info = data.map(off, self.Type_Parameter_Target)
            off += self.target_info.get_size()
        elif self.target_type['value'] == 0x10:
            self.target_info = data.map(off, self.SuperType_Target)
            off += self.target_info.get_size()
        elif self.target_type['value'] in (0x11, 0x12):
            self.target_info = data.map(off, self.TypeParameter_BoundTarget)
            off += self.target_info.get_size()
        elif self.target_type['value'] == 0x16:
            self.target_info = data.map(off, self.Formal_Parameter_Type)
            off += self.target_info.get_size()
        elif self.target_type['value'] == 0x17:
            self.target_info = data.map(off, self.Throws_Target)
            off += self.target_info.get_size()
        elif self.target_type['value'] in (0x40, 0x41):
            self.target_info = data.map(off, self.Local_Var_Target)
            off += self.target_info.get_size()
        elif self.target_type['value'] == 0x42:
            self.target_info = data.map(off, self.Catch_Target)
            off += self.target_info.get_size()
        elif self.target_type['value'] in range(0x43, 0x47):
            self.target_info = data.map(off, self.Offset_Target)
            off += self.target_info.get_size()
        elif self.target_type['value'] in range(0x47, 0x4c):
            self.target_info = data.map(off, self.Type_Argument_Target)
            off += self.target_info.get_size()
        self.target_path = data.map(off, self.Type_Path)
        off += self.target_path.get_size()
        self.end = data.map(pff, self.Type_Annot_End_Internal)
    
    def get_size(self):
        i = 1 if self.target_info == None else self.target_info.get_size() + 1
        return i + self.target_path.get_size() + self.end.get_size() 

    def pretty_printer(self, before='', end='\n'):
        print(before + "TARGET_TYPE : " + str(self.target_type['value']),
                end=end)
        before += "\t"
        print(before + "TARGET_INFO : " 
                + ("(Empty)" if self.target_info == None else ""), end=end)
        if self.target_info != None:
            self.target_info.pretty_printer(before=before + "\t", end=end)
        self.target_path.pretty_printer(before=before, end=end)
        self.end.pretty_printer(before=before, end=end)


# TODO Test
# Needed for the CL_Annotation
class Element_value(sm.Struct):
    tag = sf.IntField("Tag") # on 1 byte only

    class Const_Val_Internal(sm.Struct):
        const_value_index = sf.IntField("Const Value Index",
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)

        def get_size(self):
            return 2

        def pretty_printer(self, before="", end="\n"):
            print(before + "CONST_VALUE_INDEX : " +
                    str(self.const_value_index['value']), end=end)

    class Enum_Const_Internal(sm.Struct):
        type_name_index = sf.IntField("Type Name Index",
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)
        const_name_index = sf.IntField("Const Name Index",
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)

        def get_size(self):
            return 4

        def pretty_printer(self, before="", end="\n"):
            print(before + "TYPE_NAME_INDEX : " +
                    str(self.type_name_index['value']), end=end)
            print(before + "CONST_NAME_INDEX : " +
                    str(self.const_name_index['value']), end=end)

    class Class_Info_Internal(sm.Struct):
        class_info_index = sf.IntField("Class Info Index",
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)

        def get_size(self):
            return 2

        def pretty_printer(self, before="", end="\n"):
            print(before + "CLASS_INFO_INDEX : " +
                    str(self.class_info_index['value']), end=end)

    class Array_Value_Internal(sm.Struct):
        num_values = sf.IntField("Class Info Index",
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)

        def _setup(self, data):
            self.values = []
            off = self._srddl._offset + 2
            for _ in range(int(self.num_values['value'])):
                self.values.append(data.map(off, Element_value))
                off += self.values[-1].get_size()

        def get_size(self):
            i = 2
            for e in self.values:
                i += e.get_size()
            return i

        def pretty_printer(self, before="", end="\n"):
            print(before + "NUMBER_VALUE : " + str(self.num_values['value']),
                    end=end)
            if len(self.values) == 0:
                print(before + "VALUES : (Empty)", end=end)
            else:
                print(before + "VALUES :", end=end)

            for i in range(len(self.values)):
                self.values[i].pretty_printer(before=before + "\t#" + str(i)
                        + " ", end=end)

    def _setup(self, data):
        offset = self._srddl._offset + 1
        if self.tag['value'] in [ord(i) for i in "BCDFIJSZs"]:
            self.value = data.map(offset, Const_Val_Internal)
        elif self.tag['value'] == ord("e"):
            self.value = data.map(offset, Enum_Const_Internal)
        elif self.tag['value'] == ord("c"):
            self.value = data.map(offset, Class_Info_Internal)
        elif self.tag['value'] == ord("@"):
            self.value = data.map(offset, CL_Annotation)
        elif self.tag['value'] == ord("["):
            self.value = data.map(offset, Array_Value_Internal)

    def get_size(self):
        return 1 + self.value.get_size()

    def pretty_printer(self, before="", end="\n"):
        s = ""
        if self.tag['value'] == ord("B"):
            s += "byte"
        if self.tag['value'] == ord("C"):
            s += "char"
        if self.tag['value'] == ord("D"):
            s += "double"
        if self.tag['value'] == ord("F"):
            s += "float"
        if self.tag['value'] == ord("I"):
            s += "int"
        if self.tag['value'] == ord("J"):
            s += "long"
        if self.tag['value'] == ord("S"):
            s += "short"
        if self.tag['value'] == ord("Z"):
            s += "boolean"
        if self.tag['value'] == ord("s"):
            s += "string"
        if self.tag['value'] == ord("e"):
            s += "enum"
        if self.tag['value'] == ord("c"):
            s += "class"
        if self.tag['value'] == ord("@"):
            s += "annotation"
        if self.tag['value'] == ord("["):
            s += "array"
        print(before + "TAG : " + str(self.tag['value']) + " (" + s + ")",
                end=end)
        self.value.pretty_printer(before=before + "\t", end=end)



# TODO Test
class CL_Annotation(sm.Struct):
    """
        u2 type_index;
        u2 num_element_value_pairs;
        {
            u2            element_name_index;
            element_value value;
        } element_value_pairs[num_element_value_pairs];
    """
    type_index = sf.IntField("Type Index", size=sf.IntField.Size.INT16,
            endianess=sf.IntField.Endianess.BIG)
    num_element_value_pairs = sf.IntField("Number of element value pairs",
            size=sf.IntField.Size.INT16,
            endianess=sf.IntField.Endianess.BIG)

    class Element_value_pairs(sm.Struct):
        element_name_index = sf.IntField("Element Name Index",
                size=sf.IntField.Size.INT16,
                endianess=sf.IntField.Endianess.BIG)

        def _setup(self, data):
            self.value = data.map(self._srddl._offset + 2, Element_value)
        def get_size(self):
            return 2 + self.value.get_size()
        def pretty_printer(self, before="", end="\n"):
            print(before + "ELT_NAME_INDEX : " +
                    str(self.element_name_index['value']), end=end)
            self.value.pretty_printer(before=before + "\t", end=end)

    def _setup(self, data):
        off = self._srddl._offset + 4
        self.element_value_pairs = []
        for _ in range(int(self.num_element_value_pairs['value'])):
            self.element_value_pairs.append(data.map(off,
                self.Element_value_pairs))
            off += self.element_value_pairs[-1].get_size()

    def get_size(self):
        i = 4
        for e in selfelement_value_pairs:
            i += e.get_size()
        return i

    def pretty_printer(self, before="", end="\n"):
        print(before + "TYPE_INDEX : " + str(self.type_index['value']), end=end)
        before += "\t"
        print(before + "NUM_ELEMENT_VALUE_PAIRS : " +
                str(self.num_element_value_pairs['value']), end=end)
        if (self.element_value_pairs == []):
            print(before + "ELT_VALUE_PAIRS : (Empty)", end=end)
        else:
            print(before + "ELT_VALUE_PAIRS :", end=end)
        for i in range(len(self.element_value_pairs)):
            self.element_value_pairs[i].pretty_printer(before=before + "\t#" +
                    str(i) + " ", end=end)



    # TODO desassembleur !

# TEST
import srddl.data as sd
d = sd.FileData(sys.argv[1], mode=sd.FileData.Mode.RDWR)
c = class_java(d, 0)
c._setup(d)
c.pretty_printer()

